import pytest
from genologics.entities import Artifact, Process

from cg_lims.EPPs.udf.copy.measurement_to_analyte import copy_udfs_to_all_analytes
from cg_lims.get.artifacts import get_artifacts, get_latest_analyte
from cg_lims.exceptions import MissingUDFsError
from tests.conftest import server


def test_copy_udf(lims):
    # GIVEN a Reception control WGS process with sample measurements which are connected to analytes, measurements
    # have barcodes stored in UDF "Output Container Barcode"
    server("reception_control_wgs")
    process = Process(lims, id="24-349794")
    measurements = get_artifacts(process=process, measurement=True)
    analytes = get_artifacts(process=process, input=True)
    barcodes = {
        "ACC10237A2": "barcode1",
        "ACC10242A8": "barcode2",
        "ACC10243A1": "barcode3",
    }
    udfs = [("Output Container Barcode", "Output Container Barcode")]
    for measurement in measurements:
        measurement.udf["Output Container Barcode"] = barcodes[measurement.samples[0].id]

    # WHEN running copy_udfs_to_all_analytes to copy the barcodes
    copy_udfs_to_all_analytes(
        measurements=measurements,
        udfs=udfs,
    )

    # THEN the correct source UDFs are copied over to the correct analytes
    for analyte in analytes:
        assert analyte.udf["Output Container Barcode"] == barcodes[analyte.samples[0].id]


def test_copy_udf_missing_udfs(lims):
    # GIVEN a Reception control WGS process with sample measurements which are connected to analytes, measurements
    # don't have values stored in the required UDF
    server("reception_control_wgs")
    process = Process(lims, id="24-349794")
    measurements = get_artifacts(process=process, measurement=True)
    udfs = [("Output Container Barcode", "Output Container Barcode")]

    # WHEN when running copy_udfs_to_all_analytes
    # THEN MissingUDFsError is raised
    with pytest.raises(MissingUDFsError) as error_message:
        copy_udfs_to_all_analytes(
            measurements=measurements,
            udfs=udfs,
        )



import pytest

from genologics.entities import Artifact, Process
from cg_lims.exceptions import InvalidValueError, MissingValueError
from cg_lims.EPPs.udf.set.set_barcode import get_barcode_set_udf
from cg_lims.get.fields import get_barcode
from cg_lims.get.artifacts import get_artifacts
from tests.conftest import server


def test_tube_barcode(lims):
    # GIVEN four artifacts, only two of them ACC9553A3PA1
    # and ACC9621A7PA1 have container type "Tube" and are the
    # only ones to be have assigned a barcode udf.

    barcode_udf = 'Output Container Barcode'
    container = 'Tube'
    server("hamilton_normalization_file")
    artifacts = [
        Artifact(lims, id="ACC9553A3PA1"),
        Artifact(lims, id="ACC9551A97PA1"),
        Artifact(lims, id="ACC9621A7PA1"),
        Artifact(lims, id="ACC9551A105PA1"),
    ]

    for artifact in artifacts:
        del artifact.udf[barcode_udf]

    # WHEN running get_barcode_set_udf
    get_barcode_set_udf(artifacts=artifacts, artifact_udf=barcode_udf, container_type=container, measurement=False)

    # THEN only tubes should get barcodes and be correct.
    for artifact in artifacts:
        if artifact.container.type.name != container:
            with pytest.raises(KeyError) as e_info:
                artifact.udf[barcode_udf]
        else:
            barcode = get_barcode(artifact)
            assert artifact.udf[barcode_udf] == barcode


def test_plate_barcode(lims):
    # GIVEN four artifacts, only two of them ACC9551A97PA1
    # and ACC9551A105PA1 have container type '96 well plate' and are the
    # only ones to be included in the final file. None of them
    # have barcode_udf. 

    barcode_udf = 'Output Container Barcode'
    container = 'Some container name'
    server("hamilton_normalization_file")
    artifacts = [
        Artifact(lims, id="ACC9553A3PA1"),
        Artifact(lims, id="ACC9551A97PA1"),
        Artifact(lims, id="ACC9621A7PA1"),
        Artifact(lims, id="ACC9551A105PA1"),
    ]

    # WHEN running get_barcode_set_udf
    with pytest.raises(MissingValueError) as error_message:
        get_barcode_set_udf(artifacts=artifacts, artifact_udf=barcode_udf, container_type=container, measurement=False)

    # THEN InvalidValueError exception should be raised.
    # Because plate barcode has no specific barcode.
    assert(
        f"No barcode assigned."
        in error_message.value.message
    )


def test_pool_barcode(lims):
    # GIVEN artifacts with multiple samples - a pool -
    # and container type 'Tube'.
    server("pool_samples_TWIST")
    barcode_udf = 'Output Container Barcode'
    container = 'Tube'
    process = Process(lims, id="122-202801")
    artifacts = get_artifacts(process=process, input=False)
    
    for artifact in artifacts:
        if len(artifact.samples) == 1:
            exit()

    # WHEN running function get_barcode_set_udf.
    get_barcode_set_udf(artifacts=artifacts, artifact_udf=barcode_udf, container_type=container, measurement=False)

    # THEN correct barcode should be assigned.
    for artifact in artifacts:        
        barcode = get_barcode(artifact)
        assert artifact.udf[barcode_udf] == barcode


def test_no_container_type(lims):
    # GIVEN four artifacts, two of them (ACC9553A3PA1
    # and ACC9621A7PA1) have container type "Tube" and
    # the other two (ACC9551A97PA1 and ACC9551A105PA1)
    # belong to a "96 well plate" container

    barcode_udf = 'Output Container Barcode'
    server("hamilton_normalization_file")
    artifacts = [
        Artifact(lims, id="ACC9553A3PA1"),
        Artifact(lims, id="ACC9551A97PA1"),
        Artifact(lims, id="ACC9621A7PA1"),
        Artifact(lims, id="ACC9551A105PA1"),
    ]

    for artifact in artifacts:
        del artifact.udf[barcode_udf]

    # WHEN running get_barcode_set_udf
    get_barcode_set_udf(artifacts=artifacts, artifact_udf=barcode_udf, container_type="", measurement=False)

    # THEN all artifacts should get barcodes and be correct.
    for artifact in artifacts:
        barcode = get_barcode(artifact)
        assert artifact.udf[barcode_udf] == barcode


def test_invalid_value(lims):
    # GIVEN artifacts missing container type.
    server("missing_reads_pool")
    barcode_udf = 'Output Container Barcode'
    container = "Tube"
    artifact = [Artifact(lims, id="2-1439512")]

    # WHEN running function get_barcode_set_udf.
    with pytest.raises(InvalidValueError) as error_message:
        get_barcode_set_udf(artifacts=artifact, artifact_udf=barcode_udf, container_type=container, measurement=False)

    # THEN InvalidValueError should be triggered.
    assert (
        f"Samples ACC7239A44, are missing udf container or udf "
        in error_message.value.message
    )


def test_missing_value(lims):
    # GIVEN two artifacts, with container "Tube"
    # and container set to "No_tubes".
    barcode_udf = 'Output Container Barcode'
    container = 'No_tubes'
    server("hamilton_normalization_file")
    artifacts = [
        Artifact(lims, id="ACC9553A3PA1"),
        Artifact(lims, id="ACC9621A7PA1"),
    ]

    for artifact in artifacts:
        del artifact.udf[barcode_udf]

    # WHEN running function get_barcode_set_udf.
    with pytest.raises(MissingValueError) as error_message:
        get_barcode_set_udf(artifacts=artifacts, artifact_udf=barcode_udf, container_type=container, measurement=False)

    # THEN no barcodes should be assigned and barcode_udf should not 
    # exists.
    assert (
        f"No barcode assigned. Check parameters."
        in error_message.value.message
    )
    
    with pytest.raises(KeyError):
        artifact.udf[barcode_udf]


def test_container_on_measurement(lims):
    # GIVEN measurements with different containers.
    barcode_udf = 'Output Container Barcode'
    container = 'Tube'
    server("reception_control_wgs")
    process = Process(lims, id="24-349794")
    measurements = get_artifacts(process=process, measurement=True)

    for measurement in measurements:
        try:
            del measurement.udf[barcode_udf]
        except:
            continue

    # WHEN running get_barcode_set_udf
    get_barcode_set_udf(artifacts=measurements, artifact_udf=barcode_udf, container_type=container, measurement=True)

    # THEN barcodes should be assigned on measurement level only to samples with correct container.
    for measurement in measurements:
        if measurement.samples[0].artifact.container.type.name != container:
            with pytest.raises(KeyError):
                measurement.udf[barcode_udf]
        
        else:
            barcode = get_barcode(measurement.samples[0].artifact)
            assert measurement.udf[barcode_udf] == barcode


def test_missing_value_on_measurement(lims):
    # GIVEN measurements with different containers.
    barcode_udf = 'Output Container Barcode'
    container = 'foobar'
    server("reception_control_wgs")
    process = Process(lims, id="24-349794")
    measurements = get_artifacts(process=process, measurement=True)

    for measurement in measurements:
        try:
            del measurement.udf[barcode_udf]
        except:
            continue

    # WHEN running function get_barcode_set_udf.
    with pytest.raises(MissingValueError) as error_message:
        get_barcode_set_udf(artifacts=measurements, artifact_udf=barcode_udf, container_type=container, measurement=True)

    # THEN no barcodes should be assigned and barcode_udf should not 
    # exists.
    assert (
        f"No barcode assigned. Check parameters."
        in error_message.value.message
    )
    
    with pytest.raises(KeyError):
        measurement.udf[barcode_udf]



import logging
import sys
import click

from typing import List
from genologics.lims import Artifact
from cg_lims import options
from cg_lims.exceptions import LimsError, InvalidValueError, MissingValueError
from cg_lims.get.artifacts import get_artifacts
from cg_lims.get.fields import get_barcode

LOG = logging.getLogger(__name__)


def get_barcode_set_udf(
    artifacts: List[Artifact],
    artifact_udf: str,
    container_type: str,
    measurement: bool,
):

    """Assigning barcode to a barcode udf.
    
    - Assigned on container type if selected.
    - Gives warning if no samples assigned.
    - Checks if script failed and what the cause might be.
    - Informs user if sample doesn't have a defined barcode.

    """
    
    assigned_artifacts = 0
    failed_samples = []

    for artifact in artifacts:
        if measurement:
            if len(artifact.samples) > 1:
                LOG.info(f"Sample {str(artifact.id)} is a pool, and will be excluded")
            
            else:
                measurement_artifact = artifact
                artifact = artifact.input_artifact_list()[0]

        try:
            art_container_type = artifact.container.type.name

            # Assigns barcodes too all samples if no container type is given
            if not container_type:
                barcode = get_barcode(artifact)
                artifact.udf[artifact_udf] = barcode
                artifact.put()
                assigned_artifacts += 1

            # Removes all incorrect container types if option is used.
            elif art_container_type.lower() != container_type.lower():
                LOG.info(
                    f"Sample {str(artifact.samples[0].id)} does not have container"
                    f" type {container_type} therefore excluded."
                )

            # Sample has the correct "container_type".
            else:
                barcode = get_barcode(artifact)

                if measurement:
                    measurement_artifact.udf[artifact_udf] = barcode
                    measurement_artifact.put()
                    assigned_artifacts += 1
                
                else:
                    artifact.udf[artifact_udf] = barcode
                    artifact.put()
                    assigned_artifacts += 1

        # Collects failed samples.
        except:
            failed_samples.append(artifact.samples[0].id)

    # Triggered if any failed samples.
    if failed_samples:
        failed_message = ", ".join(map(str, failed_samples))
        raise InvalidValueError(
            f"Samples {failed_message}, are missing udf container or udf \"{artifact_udf}\"."
        )
    
    # Notifies that no sample got a barcode.
    elif not assigned_artifacts:
        raise MissingValueError(f"No barcode assigned. Check parameters.")


@click.command()
@options.artifact_udf(help="The name of the barcode udf.")
@options.input()
@options.measurement()
@options.container_type()
@click.pass_context
def assign_barcode(
    ctx: click.Context,
    artifact_udf: str,
    input: bool,
    measurement: bool,
    container_type: str,
):
    """Assigned barcode to UDF"""

    LOG.info(f"Running {ctx.command_path} with params: {ctx.params}")
    process = ctx.obj["process"]
    artifacts = get_artifacts(process=process, input=input, measurement=measurement)
    try:
        get_barcode_set_udf(
            artifacts=artifacts,
            artifact_udf=artifact_udf,
            container_type=container_type,
            measurement=measurement,
        )
        message = "Barcodes were successfully generated."
        LOG.info(message)
        click.echo(message)
    except LimsError as e:
        LOG.error(e.message)
        sys.exit(e.message)



from tests.conftest import server
from pathlib import Path
from cg_lims.exceptions import InvalidValueError, MissingValueError
import pytest

from genologics.entities import Artifact
from cg_lims.EPPs.files.barcode_tubes import get_data_and_write


def test_with_diff_containers(barcode_tubes_csv, lims):
    # GIVEN four artifacts, only two of them ACC9553A3PA1
    # and ACC9621A7PA1 have container type "Tube" and are the
    # only ones to be included in the final file.

    server("hamilton_normalization_file")
    file = Path("file_name_1")
    artifacts = [
        Artifact(lims, id="ACC9553A3PA1"),
        Artifact(lims, id="ACC9551A97PA1"),
        Artifact(lims, id="ACC9621A7PA1"),
        Artifact(lims, id="ACC9551A105PA1"),
    ]

    # WHEN running get_data_and_write
    get_data_and_write(artifacts=artifacts, file=file)

    # THEN the file has been created and the file content is as expected.
    file_content = file.read_text()
    file.unlink()
    assert file_content == barcode_tubes_csv


def test_MissingValueError(lims):
    # GIVEN artifacts without container type "Tube".
    server("hamilton_normalization_file")
    file = Path("file_name_2")
    artifacts = [
        Artifact(lims, id="2-2463481"),
        Artifact(lims, id="2-2463482"),
        Artifact(lims, id="2-2463483"),
        Artifact(lims, id="2-2463484"),
        Artifact(lims, id="2-2463485"),
    ]

    # WHEN running get_data_and_write
    with pytest.raises(MissingValueError) as error_message:
        get_data_and_write(artifacts, file=file)

    # Then InvalidValueError exception should be raised, and no file created.
    assert(
        f"Missing samples with container type \"Tube\"."
        in error_message.value.message
    )
    assert file.exists() is False


def test_InvalidValueError(lims):
    # GIVEN artifacts without container type.
    server("flat_tests")
    file = Path("file_name_2")
    artifacts = [
        Artifact(lims, id="1"),
        Artifact(lims, id="2"),
    ]

    # WHEN running get_data_and_write
    with pytest.raises(InvalidValueError) as error_message:
        get_data_and_write(artifacts, file=file)

    # Then InvalidValueError exception should be raised, and no file created.
    assert(
        f"The following samples are missing a container: S1 S1"
        in error_message.value.message
    )
    assert file.exists() is False



import click
import csv
import logging
from genologics.lims import Lims
from genologics.entities import Sample

from cg_lims.exceptions import LimsError
from cg_lims.get.artifacts import get_sample_artifact
from typing import List

LOG = logging.getLogger(__name__)


def check_cancelled_status(sample: Sample) -> bool:
    status = sample.udf.get("cancelled")
    if status is None or status.lower() != "yes":
        return False
    else:
        return True


def get_samples_from_source(source_types: List[str], skip_cancelled: bool, lims: Lims) -> List[Sample]:
    samples_from_sources = []
    for source in source_types:
        udf_dict = {"Source": source}
        found_samples = lims.get_samples(udf=udf_dict)
        cancelled_samples = 0
        for sample in found_samples:
            if skip_cancelled and check_cancelled_status(sample=sample):
                cancelled_samples += 1
                message = f"Skipping cancelled sample {sample.id}."
                LOG.info(message)
                continue
            samples_from_sources.append(sample)
        message = f"Found {len(found_samples)} sample(s) of source type '{source}'."
        if skip_cancelled:
            message = message + f" Skipped {cancelled_samples} of them because of cancellation status."
        LOG.info(message)
        click.echo(message)
    return samples_from_sources


def print_fetched_samples(
        samples: List[Sample],
        udf_name: str,
        max_amount: str,
        min_amount: str,
        file_path: str,
        lims: Lims,
) -> None:

    sample_rows = []
    missing_udf = 0
    outside_threshold = 0
    for sample in samples:
        sample_artifact = get_sample_artifact(lims=lims, sample=sample)
        if not udf_name:
            sample_rows.append([sample.id, sample.udf.get("Source")])
            continue
        udf_value = sample_artifact.udf.get(udf_name)
        if not udf_value:
            missing_udf += 1
            continue
        if min_amount and float(udf_value) < float(min_amount):
            outside_threshold += 1
            continue
        if max_amount and float(udf_value) > float(max_amount):
            outside_threshold += 1
            continue
        sample_rows.append([sample.id, udf_value, sample.udf.get("Source")])

    if not udf_name:
        headers = ["Sample ID", "Source"]
    else:
        headers = ["Sample ID", udf_name, "Source"]
        if outside_threshold > 0 or missing_udf > 0:
            message = f"Removed a total of {outside_threshold} sample(s) that did not fulfill threshold requirements" \
                      f" and {missing_udf} because of missing UDF values."
            LOG.info(message)
            click.echo(message)
    with open(f"{file_path}.csv", "w", newline="\n") as new_csv:
        wr = csv.writer(new_csv, delimiter=",")
        wr.writerow(headers)
        wr.writerows(sample_rows)


@click.command()
@click.option(
    "--source-type",
    "-s",
    required=False,
    multiple=True,
    help="What source type should be fetched. Allows multiple different sources.",
)
@click.option(
    "--max-udf-amount",
    "-mx",
    required=False,
    help="Don't fetch samples OVER a certain UDF value.",
)
@click.option(
    "--min-udf-amount",
    "-mn",
    required=False,
    help="Don't fetch samples BELOW a certain UDF value.",
)
@click.option(
    "--udf-name",
    "-u",
    required=False,
    help="Fetch samples depending on a specific UDF value. "
         "Use in conjecture with max and/or min value options. "
         "At the moment only works on Sample Artifact UDFs.",
)
@click.option(
    "--file-name",
    "-f",
    required=True,
    help="Prefix of output sample csv-file.",
)
@click.option(
    "--skip-cancelled",
    required=False,
    is_flag=True,
    help="Flag to skip samples which have been cancelled in LIMS.",
)
@click.pass_context
def fetch_sample_info(
        ctx,
        source_type: List[str],
        max_udf_amount: str,
        min_udf_amount: str,
        udf_name: str,
        file_name: str,
        skip_cancelled: bool = False,
):
    LOG.info(f"Running {ctx.command_path} with params: {ctx.params}")

    lims = ctx.obj["lims"]

    try:
        samples_from_sources = get_samples_from_source(
            source_types=source_type,
            skip_cancelled=skip_cancelled,
            lims=lims,
        )
        print_fetched_samples(
            samples=samples_from_sources,
            udf_name=udf_name,
            max_amount=max_udf_amount,
            min_amount=min_udf_amount,
            file_path=file_name,
            lims=lims,
        )
        message = f"Sample info file has been created."
        LOG.info(message)
        click.echo(message)
    except LimsError as e:
        LOG.error(e.message)
        click.echo(e.message)



import re
import logging
import sys
import click
import pandas as pd

from pathlib import Path
from typing import List
from genologics.lims import Artifact
from cg_lims import options
from cg_lims.exceptions import LimsError, InvalidValueError, MissingValueError
from cg_lims.get.artifacts import get_artifacts
from cg_lims.get.fields import get_barcode


LOG = logging.getLogger(__name__)

def get_data_and_write(
    artifacts: List[Artifact], 
    file: str,
):
    """Making a barcode csv file with sample names as barcode."""

    file_rows = []
    unexpected_container_type = 0
    failed_samples = []

    for artifact in artifacts:
        try:
            barcode = get_barcode(artifact)
            art_container_type = artifact.container.type.name

            if art_container_type.lower() != "tube":

                LOG.info(
                    f"Sample {barcode} does not have container type \"Tube\" therefore excluded."
                )
                continue
            
            
            #Function to get project nr and sample id
            f=lambda x: re.findall(r'\d+',x)
            
            file_rows.append([barcode, f(barcode)[0], int(f(barcode)[1])])
            #                           ^^^^^          ^^^^^^
            #  e.g.     ACC10001A1 =    10001             1 
        
        except:
            unexpected_container_type = + 1
            failed_samples.append(artifact.samples[0].id)

    if unexpected_container_type:
        failed_message = " ".join(map(str, failed_samples))
        raise InvalidValueError(
            f"The following samples are missing a container: {failed_message} of type \"Tube\"."
        )

    if file_rows == []:
        raise MissingValueError(
            f"Missing samples with container type \"Tube\"."
        )

    # Sort and create csv file.
    unfiltered_df = pd.DataFrame(
        data=file_rows, 
        index = None, 
        columns = [
            "Barcode",
            "Project",
            "id",
        ]
        ).sort_values(by=["Project","id"])
    
    # Filter off unwanted columns
    out = unfiltered_df["Barcode"]
    out.to_csv(Path(file), index=False)
    



@click.command()
@options.file_placeholder(help="Barcode Tubes")
@options.input()
@click.pass_context
def make_barcode_csv(
    ctx: click.Context, file: str, input: bool,
):
    """Script to make barcode for tubes"""

    LOG.info(f"Running {ctx.command_path} with params: {ctx.params}")
    process = ctx.obj["process"]
    artifacts = get_artifacts(process=process, input=input)
    try:
        get_data_and_write(
            artifacts=artifacts,
            file=f"{file}-Barcode-Tubes.csv",
        )
        click.echo("The file was successfully generated.")
    except LimsError as e:
        sys.exit(e.message)



from typing import List, Tuple
from genologics.entities import Artifact

from cg_lims import options
from cg_lims.exceptions import LimsError, MissingUDFsError
from cg_lims.get.artifacts import get_artifacts
from cg_lims.set.udfs import copy_artifact_to_artifact

import logging
import sys
import click

LOG = logging.getLogger(__name__)


def copy_udfs_to_all_analytes(
    measurements: List[Artifact],
    udfs: List[Tuple[str, str]],
):
    """Looping over all analytes and copying over the measurement values for the given UDFs."""

    failed_artifacts = 0
    for measurement in measurements:
        try:
            analyte = measurement.input_artifact_list()[0]
            copy_artifact_to_artifact(
                destination_artifact=analyte,
                source_artifact=measurement,
                artifact_udfs=udfs,
            )
        except:
            failed_artifacts += 1
    if failed_artifacts:
        raise MissingUDFsError(
            message=f"Failed to set artifact udfs on {failed_artifacts} artifacts. See log for details"
        )


@click.command()
@options.artifact_udfs(
    help="The name of the udf that you want to set"
)
@click.pass_context
def measurement_to_analyte(
    ctx,
    artifact_udfs: List[str],
):
    """
    Script to copy UDFs to measurements in the current step, from their source analytes in the same step.
    """

    LOG.info(f"Running {ctx.command_path} with params: {ctx.params}")
    process = ctx.obj["process"]

    udfs = list(zip(artifact_udfs, artifact_udfs))

    try:
        measurements = get_artifacts(process=process, measurement=True)
        copy_udfs_to_all_analytes(
            measurements=measurements,
            udfs=udfs,
        )
        click.echo("Udfs have been set on all samples.")
    except LimsError as e:
        sys.exit(e.message)
